from typing import List

from .. import ExtensionMixin

from ...flarum.core.users import UserFromBulk
from ...flarum.core.posts import Post, PostFromBulk, PostFromNotification
from ...error_handler import parse_request


AUTHOR = 'flarum'
NAME = 'likes'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []


class LikesPostFromNotificationMixin(PostFromNotification):
    @property
    def canLike(self) -> bool:
        return self.attributes.get("canLike", False)
    

    def __like_or_unlike(self, liked: bool=True):
        """
            A function to either like or unlike post, to prevent repetition.
        """

        patch_data = {
            "data": {
                "type": "posts",
                "id": self.id,
                "attributes": {
                    "isLiked": liked
                }
            }
        }

        raw = self.user.session.patch(f"{self.user.api_urls['posts']}/{self.id}", json=patch_data)
        json = parse_request(raw)

        return Post(user=self.user, _fetched_data=json)


    def like(self):
        """
            Likes a post.
        """

        return self.__like_or_unlike(liked=True)


    def unlike(self):
        """
            Unlikes liked post.
        """

        return self.__like_or_unlike(liked=False)



class LikesPostFromBulkMixin(PostFromBulk, LikesPostFromNotificationMixin):
    def get_liked_by(self):
        all_users = list() # type: List[UserFromBulk]

        for raw_user in self.relationships.get("likes", {}).get("data", [{}]):
            if raw_user.get("type", None) == "users":
                id_to_find = raw_user.get("id", None)

                if id_to_find:
                    for possible_raw_user in self._parent_included:
                        if (possible_raw_user.get("type", None) == 'users') and (possible_raw_user.get("id", None) == id_to_find):
                            user = UserFromBulk(user=self.user, _fetched_data=dict(data=possible_raw_user))

                            all_users.append(user)

        return all_users



class LikesExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, PostFromNotification, LikesPostFromNotificationMixin)
        super().mixin(self, PostFromBulk, LikesPostFromBulkMixin)



from typing import Literal, Optional, List


class Filter():
    def __init__(self, order_by: Optional[Literal['commentCount', '-commentCount', 'createdAt', '-createdAt']]=None, query: Optional[str]=None, limit: int=20, page: Optional[int]=None, include: Optional[List[str]]=None, additional_data: Optional[dict]=None):
        self.order_by = order_by
        self.query = query
        self.limit = limit
        self.page = page
        self.include = include
        self.additional_data = additional_data


    @property
    def to_dict(self):
        __data = {}

        if self.order_by:
            __data["sort"] = self.order_by

        if self.query:
            __data["filter[q]"] = self.query

        if self.limit:
            __data["page[limit]"] = self.limit

        if self.page:
            __data["page[offset]"] = self.page * self.limit

        if self.include:
            __data["include"] = ','.join(self.include)


        if self.additional_data:
            __data.update(self.additional_data)
        
        return __data



from pyflarum.flarum.core.discussions import DiscussionFromNotification
from typing import TYPE_CHECKING, Optional, List

from datetime import datetime

# Avoid my greatest enemy in Python: circular import:
if TYPE_CHECKING:
    from ...session import FlarumUser

from ...flarum.core.users import UserFromNotification
from ...flarum.core.posts import PostFromNotification
from ...error_handler import parse_request
from ...datetime_conversions import flarum_to_datetime


class Notifications(dict):
    """
        A data of multiple notifications fetched from the API.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user

        super().__init__(_fetched_data)


    def __iter__(self):
        return iter(self.get_notifications())


    @property
    def links(self) -> dict:
        return self.get("links", {})


    @property
    def first_link(self) -> Optional[str]:
        return self.links.get("first", None)


    @property
    def previous_link(self) -> Optional[str]:
        return self.links.get("prev", None)


    @property
    def next_link(self) -> Optional[str]:
        return self.links.get("next", None)


    @property
    def data(self) -> List[dict]:
        return self.get("data", [{}])


    @property
    def included(self) -> List[dict]:
        return self.get("included", [{}])


    def get_notifications(self):
        """
            All notifications from the `Notifications` object.
        """

        all_notifications = [] # type: List[Notification]

        for raw_notification in self.data:
            if raw_notification.get("type", None) == 'notifications':
                notification = Notification(user=self.user, _fetched_data=dict(data=raw_notification, _parent_included=self.included))
                all_notifications.append(notification)

        return all_notifications


    def mark_all_as_read(self) -> bool:
        raw = self.user.session.post(f"{self.user.api_urls['notifications']}/read")
        parse_request(raw)

        return True



class Notification(dict):
    """
        Notification, that always has properties defined.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user
        super().__init__(_fetched_data)


    @property
    def data(self) -> dict:
        return self.get("data", {})


    @property
    def type(self) -> Optional[str]:
        return self.data.get("type", None)


    @property
    def id(self) -> Optional[int]:
        raw = self.data.get("id", None)

        if raw:
            return int(raw)


    @property
    def attributes(self) -> dict:
        return self.data.get("attributes", {})


    @property
    def contentType(self) -> Optional[str]:
        return self.attributes.get("contentType", None)


    @property
    def content(self) -> Optional[dict]:
        return self.attributes.get("content", None)


    @property
    def new_post_number(self) -> Optional[int]:
        if self.content and self.contentType == "newPost":
            post_number = self.content.get("postNumber", None)

            if post_number:
                return int(post_number)


    @property
    def reply_number(self) -> Optional[int]:
        if self.content and self.contentType == "postMentioned":
            reply_number = self.content.get("replyNumber", None)

            if reply_number:
                return int(reply_number)


    @property
    def createdAt(self) -> Optional[datetime]:
        raw = self.attributes.get("createdAt", None)

        return flarum_to_datetime(raw)


    @property
    def isRead(self) -> bool:
        return self.attributes.get("isRead", False)


    @property
    def relationships(self) -> dict:
        return self.data.get("relationships", {})


    @property
    def _parent_included(self) -> List[dict]:
        return self.get("_parent_included", [{}])


    def from_user(self):
        id = self.relationships.get("fromUser", {}).get("data", {}).get("id", None)
        
        for raw_user in self._parent_included:
            if raw_user.get("id", None) == id and raw_user.get("type", None) == 'users':
                user = UserFromNotification(user=self.user, _fetched_data=dict(data=raw_user))
                return user

        return None


    def get_subject(self):
        id = self.relationships.get("subject", {}).get("data", {}).get("id", None)
        
        for raw_subject in self._parent_included:
            if raw_subject.get("id", None) == id:
                notification_type = raw_subject.get("type", None)

                if notification_type == 'discussions':
                    return DiscussionFromNotification(user=self.user, _fetched_data=dict(data=raw_subject))

                elif notification_type == 'posts':
                    return PostFromNotification(user=self.user, _fetched_data=dict(data=raw_subject, _parent_included=self._parent_included))

        return None


    def mark_as_read(self):
        post_data = {"is_read": True}
        raw = self.user.session.patch(f"{self.user.api_urls['notifications']}/{self.id}", json=post_data)
        parse_request(raw)

        return True



from pyflarum.flarum.core.groups import Group
from typing import TYPE_CHECKING, Optional, List

# Avoid my greatest enemy in Python: circular import:
if TYPE_CHECKING:
    from ...session import FlarumUser

from datetime import datetime

from ...error_handler import FlarumError, parse_request
from ...datetime_conversions import flarum_to_datetime


class Forum(dict):
    """
        Forum JSON data
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user

        super().__init__(_fetched_data)


    @property
    def data(self) -> dict:
        return self.get("data", {})


    @property
    def type(self) -> Optional[str]:
        return self.data.get("type", None)


    @property
    def id(self) -> Optional[int]:
        raw = self.data.get("id", None)

        if raw:
            return int(raw)


    @property
    def attributes(self) -> dict:
        return self.data.get("attributes", {})


    @property
    def title(self) -> Optional[str]:
        return self.attributes.get("title", None)


    @property
    def description(self) -> Optional[str]:
        return self.attributes.get("description", None)


    @property
    def showLanguageSelector(self) -> bool:
        return self.attributes.get("showLanguageSelector", False)


    @property
    def baseUrl(self) -> Optional[str]:
        return self.attributes.get("baseUrl", None)


    @property
    def basePath(self) -> Optional[str]:
        return self.attributes.get("basePath", None)


    @property
    def debug(self) -> bool:
        return self.attributes.get("debug", False)


    @property
    def apiUrl(self) -> Optional[str]:
        return self.attributes.get("apiUrl", None)


    @property
    def welcomeTitle(self) -> Optional[str]:
        return self.attributes.get("welcomeTitle", None)


    @property
    def welcomeMessage(self) -> Optional[str]:
        return self.attributes.get("welcomeMessage", None)


    @property
    def themePrimaryColor(self) -> Optional[str]:
        return self.attributes.get("themePrimaryColor", None)


    @property
    def themeSecondaryColor(self) -> Optional[str]:
        return self.attributes.get("themeSecondaryColor", None)


    @property
    def logoUrl(self) -> Optional[str]:
        return self.attributes.get("logoUrl", None)


    @property
    def faviconUrl(self) -> Optional[str]:
        return self.attributes.get("faviconUrl", None)


    @property
    def headerHtml(self) -> Optional[str]:
        return self.attributes.get("headerHtml", None)


    @property
    def footerHtml(self) -> Optional[str]:
        return self.attributes.get("footerHtml", None)


    @property
    def allowSignUp(self) -> bool:
        return self.attributes.get("allowSignUp", False)


    @property
    def defaultRoute(self) -> Optional[str]:
        return self.attributes.get("defaultRoute", None)


    @property
    def canViewForum(self) -> bool:
        return self.attributes.get("canViewForum", False)


    @property
    def canStartDiscussion(self) -> bool:
        return self.attributes.get("canStartDiscussion", False)


    @property
    def canSearchUsers(self) -> bool:
        return self.attributes.get("canSearchUsers", False)


    @property
    def adminUrl(self) -> Optional[str]:
        return self.attributes.get("adminUrl", None)


    @property
    def version(self) -> Optional[str]:
        return self.attributes.get("version", None)


    @property
    def allowUsernameMentionFormat(self) -> bool:
        return self.attributes.get("allowUsernameMentionFormat", False)


    @property
    def relationships(self) -> dict:
        return self.get("relationships", {})


    def get_groups(self):
        all_groups = list() # type: List[Group]

        for raw_group in self.relationships.get("groups", {}).get("data", [{}]):
            if raw_group.get("type", None) == 'groups':
                group = Group(user=self.user, _fetched_data=dict(data=raw_group))

                all_groups.append(group)

        return all_groups


    @property
    def included(self) -> List[dict]:
        return self.get("included", [{}])



from typing import Optional

from .. import ExtensionMixin

from ...flarum.core.forum import Forum
from ...flarum.core.posts import PostFromNotification


AUTHOR = 'flarum'
NAME = 'flags'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class FlagsForumMixin(Forum):
    @property
    def canViewFlags(self) -> bool:
        return self.attributes.get("canViewFlags", False)


    @property
    def flagCount(self) -> Optional[int]:
        return self.attributes.get("flagCount", None)


    @property
    def guidelinesUrl(self) -> Optional[str]:
        return self.attributes.get("guidelinesUrl", None)



class FlagsPostFromNotificationMixin(PostFromNotification):
    @property
    def canFlag(self) -> bool:
        return self.attributes.get("canFlag", False)



class FlagsExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, PostFromNotification, FlagsPostFromNotificationMixin)
        super().mixin(self, Forum, FlagsForumMixin)



from typing import Optional, List, Dict

from requests.models import Response
from json.decoder import JSONDecodeError


class FlarumError(Exception):
    """
        Generic class for all Flarum related errors.
    """

    def __init__(self, message: Optional[str]=None, status: Optional[int]=None, code: Optional[str]=None, details: Optional[str]=None):
        self.status = status
        self.code = code
        self.details = details

        return super().__init__(message)


class MissingExtensionError(Exception):
    """Missing pyFlarum extension error."""
    pass


class MissingExtensionWarning(Warning):
    """Missing pyFlarum extension warning."""
    pass


def parse_request(request: Response):
    try:
        # Includes opening and closing brackets:
        if len(request.text) >= 2:
            json = request.json() # type: dict

        else:
            json = {}

    except JSONDecodeError as e:
        json = {'errors': [{'code': 0, 'detail': e}]}


    if 'errors' in json:
        return handle_errors(json['errors'])

    elif not 200 <= request.status_code <= 204:
        return handle_errors(status_code=request.status_code)


    return json


def handle_errors(errors: Optional[List[Dict[str, str]]]=None, status_code: Optional[str]=None):
    """
        Handles Flarum & request related errors. Should be called on error only.
    """

    if errors:
        for error in errors:
            if status_code:
                status = status_code

            else:
                s = error.get('status', None)

                if s:
                    status = int(s)

                else:
                    status = 'Unknown'

            code = error.get('code', 0)
            details = error.get("detail", "No further details.")


            if status == 400:
                if code == 'invalid_parameter':
                    raise FlarumError(f'Error {status}: Invalid parameter. This usually happens when you provide an invalid filter when filtering data. {details}', status=status, code=code, details=details)

                else:
                    raise FlarumError(f'Error {status}: {code} - {details}', status=status, code=code, details=details)


            if status == 401:
                if code == 'not_authenticated':
                    raise FlarumError(f'Error {status}: Not authenticated. Please, check whether your authentication details (username & password) are correct and try again. {details}', status=status, code=code, details=details)

                else:
                    raise FlarumError(f'Error {status}: {code} - {details}', status=status, code=code, details=details)

            else:
                raise FlarumError(f'Error {status}: {code} - {details}', status=status, code=code, details=details)

    else:
        raise FlarumError(f'Request related error: {status_code}')



from typing import Literal, NoReturn, TYPE_CHECKING, Optional, Union, List

# Avoid my greatest enemy in Python: circular import:
if TYPE_CHECKING:
    from ...session import FlarumUser
    from ...flarum.core.discussions import DiscussionFromBulk, Discussion

from ...flarum.core.discussions import DiscussionFromNotification

from datetime import datetime

from ...flarum.core.users import UserFromBulk
from ...error_handler import FlarumError, parse_request
from ...datetime_conversions import flarum_to_datetime


class PreparedPost(dict):
    def __init__(self, user: 'FlarumUser', discussion: Optional[Union['Discussion', 'DiscussionFromBulk', 'DiscussionFromNotification']]=None, content: Optional[str]=None):
        self.user = user
        self.discussion = discussion
        self.content = content
    

    @property
    def to_dict(self):
        data = {
            "data": {
                "type": "posts",
                "attributes": {
                    "content": self.content
                },
                "relationships": {
                    "discussion": {
                        "data": {
                            "type": "discussions",
                            "id": self.discussion.id
                        }
                    }
                }
            }
        }

        return data


    def post(self):
        """
            Posts/creates the post. Returns the created `Post`.
        """

        raw = self.user.session.post(self.user.api_urls['posts'], json=self.to_dict)
        json = parse_request(raw)

        return Post(user=self.user, _fetched_data=json)
    create = post


class Posts(dict):
    """
        A data of multiple posts fetched from the API.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user

        super().__init__(_fetched_data)


    def __iter__(self):
        return iter(self.get_posts())


    @property
    def links(self) -> dict:
        return self.get("links", {})


    @property
    def first_link(self) -> Optional[str]:
        return self.links.get("first", None)


    @property
    def previous_link(self) -> Optional[str]:
        return self.links.get("prev", None)


    @property
    def next_link(self) -> Optional[str]:
        return self.links.get("next", None)


    @property
    def data(self) -> List[dict]:
        return self.get("data", [{}])


    @property
    def included(self) -> List[dict]:
        return self.get("included", [{}])


    def get_posts(self):
        """
            All posts from the `Posts` object.
        """

        all_posts = [] # type: List[PostFromBulk]

        for raw_post in self.data:
            if raw_post.get("type", None) == 'posts':
                post = PostFromBulk(user=self.user, _fetched_data=dict(data=raw_post, _parent_included=self.included))
                all_posts.append(post)

        return all_posts


class PostFromDiscussion(dict):
    """
        A post from `Discussion`
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user
        super().__init__(_fetched_data)



    @property
    def data(self) -> dict:
        return self.get("data", {})


    @property
    def type(self) -> Optional[str]:
        return self.data.get("type", None)


    @property
    def id(self) -> Optional[int]:
        raw = self.data.get("id", None)

        if raw:
            return int(raw)


    @property
    def attributes(self) -> dict:
        return self.data.get("attributes", {})


    @property
    def number(self) -> Optional[int]:
        return self.attributes.get("number", None)


    @property
    def createdAt(self) -> Optional[datetime]:
        raw = self.attributes.get("createdAt", None)

        return flarum_to_datetime(raw)


    @property
    def contentType(self) -> Optional[str]:
        return self.attributes.get("contentType", None)


    @property
    def contentHtml(self) -> Optional[str]:
        return self.attributes.get("contentHtml", None)



class PostFromNotification(PostFromDiscussion):
    """
        A post from `Notification`
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user
        super().__init__(user=self.user, _fetched_data=_fetched_data)



    @property
    def content(self) -> Optional[str]:
        return self.attributes.get("content", None)


    @property
    def ipAddress(self) -> Optional[str]:
        return self.attributes.get("ipAddress", None)


    @property
    def editedAt(self) -> Optional[datetime]:
        raw = self.attributes.get("editedAt", None)

        return flarum_to_datetime(raw)


    @property
    def canEdit(self) -> bool:
        return self.attributes.get("canEdit", False)


    @property
    def canDelete(self) -> bool:
        return self.attributes.get("canDelete", False)


    @property
    def canHide(self) -> bool:
        return self.attributes.get("canHide", False)


    @property
    def _parent_included(self) -> List[dict]:
        return self.get("_parent_included", [{}])


    @property
    def relationships(self) -> dict:
        return self.data.get("relationships", {})


    @property
    def url(self):
        discussion_id = self.relationships.get("discussion", {}).get("data", {}).get("id", None)

        if discussion_id:
            return f"{self.user.forum_url}/d/{discussion_id}/{self.number}"


    def __restore_or_hide(self, hide: bool, force: bool=False) -> Union[Literal[False], NoReturn]:
        if not self.canHide and not force:
            raise FlarumError(f'You do not have permission to {"hide" if hide else "unhide"} this post ({self.id}). Use `force = True` to ignore this error.')

        patch_data = {
            "data": {
                "type": "posts",
                "id": self.id,
                "attributes": {
                    "isHidden": hide
                }
            }
        }

        raw = self.user.session.patch(f"{self.user.api_urls['posts']}/{self.id}", json=patch_data)
        json = parse_request(raw)

        return Post(user=self.user, _fetched_data=json)


    def hide(self, force: bool=False):
        """
            Hides the post. Raises `FlarumError` or returns `False` if it failed, otherwise `True` is returned.
        """

        return self.__restore_or_hide(hide=True, force=force)


    def restore(self, force: bool=False):
        """
            Restores the post (unhides). Raises `FlarumError` or returns `False` if it failed, otherwise `True` is returned.
        """

        return self.__restore_or_hide(hide=False, force=force)
    unhide = restore


    def delete(self, force: bool=False):
        """
            Removes the post forever.
        """

        if not self.canDelete and not force:
            raise FlarumError(f'You do not have permission to delete this post ({self.id})')

        raw = self.user.session.delete(f"{self.user.api_urls['discussions']}/{self.id}")
        parse_request(raw)

        return True


    def get_discussion(self):
        id_to_find = self.relationships.get("discussion", {}).get("data", {}).get("id", None)

        for possible_discussion in self._parent_included:
            if (possible_discussion.get("type", None) == "discussions") and (possible_discussion.get("id", None) == id_to_find):
                discussion = DiscussionFromNotification(user=self.user, _fetched_data=dict(data=possible_discussion, _parent_included=self._parent_included))

                return discussion

        return None


    def reply_to(self, post: PreparedPost):
        """
            Replies to this `Post` with another `PreparedPost`.
        """

        to_post = post
        to_post.discussion = self.get_discussion()

        return to_post.post()


    def get_author(self):
        id = self.relationships.get("user", {}).get("data", {}).get("id", None)
        
        for raw_user in self._parent_included:
            if raw_user.get("id", None) == id and raw_user.get("type", None) == 'users':
                user = UserFromBulk(user=self.user, _fetched_data=dict(data=raw_user))

                return user

        return None


    def edit(self, new_data: PreparedPost):
        raw = self.user.session.patch(f"{self.user.api_urls['posts']}/{self.id}", json=new_data.to_dict)
        json = parse_request(raw)

        return Post(user=self.user, _fetched_data=json)



class PostFromBulk(PostFromNotification):
    """
        A post from `Posts`.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user
        super().__init__(user=self.user, _fetched_data=_fetched_data)


    @property
    def _parent_included(self) -> List[dict]:
        return self.get("_parent_included", [{}])


class Post(PostFromBulk):
    """
        A Flarum group.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user

        super().__init__(user=self.user, _fetched_data=_fetched_data)



from typing import Any, BinaryIO, Dict, List, Union, Optional, Literal
from .custom_types import AnyUser


import warnings
from datetime import datetime
from requests import Session


from .flarum.core.forum import Forum

from .flarum.core.users import MyUser, User, Users
from .flarum.core.notifications import Notifications
from .flarum.core.groups import Groups

from .flarum.core.discussions import Discussion, Discussions
from .flarum.core.posts import Post, Posts

from .flarum.core.filters import Filter


from .error_handler import MissingExtensionError, MissingExtensionWarning, parse_request
from .datetime_conversions import datetime_to_flarum


from .extensions import ExtensionMixin


class FlarumSession(object):
    def __init__(self, forum_url: str, username: Union[str]=None, password: Union[str, None]=None, api_endpoint: str="api", user_agent: str="pyflarum", session_object: Union[Session, Any]=Session()):
        self.forum_url = forum_url
        self.api_endpoint = api_endpoint
        self.username = username
        self.session = session_object

        self.session.headers.update({
            "User-Agent": user_agent
        })


        self.__authenticate(username, password)


    def __str__(self) -> str:
        return f'<{type(self).__name__}> authenticated={self.is_authenticated}; username={self.username}'


    def __repr__(self) -> str:
        return self.__str__()


    def __authenticate(self, username: Union[str]=None, password: Union[str, None]=None):
        if username and password:
            identification_data = {"identification": self.username, "password": password}

            raw = self.session.post(url=f'{self.forum_url}/api/token', json=identification_data)
            token_data = parse_request(raw)

            token = token_data.get("token", None)
            user_id = token_data.get("userId", None)

            if token:
                self.session.headers.update({
                    "Authorization": f'Token {token}; userId={user_id}',
                })

                self.is_authenticated = True

            else:
                self.is_authenticated = False

        else:
            self.is_authenticated = False


    @property
    def api_urls(self):
        """
            Simple, hardcoded `'key: value'` `dict` of Flarum's API routes for quick access.

            API routes reference (old):
            https://github.com/flarum/flarum.github.io/blob/20322c0e6011e4f304ae7e95f41594a0b086bc27/_docs/api.md
        """

        return {
            "base": f"{self.forum_url}/{self.api_endpoint}",
            "forum": f"{self.forum_url}/{self.api_endpoint}/forum",

            "settings": f"{self.forum_url}/{self.api_endpoint}/settings",
            "extensions": f"{self.forum_url}/{self.api_endpoint}/extensions",

            "discussions": f"{self.forum_url}/{self.api_endpoint}/discussions",
            "posts": f"{self.forum_url}/{self.api_endpoint}/posts",

            "users": f"{self.forum_url}/{self.api_endpoint}/users",
            "notifications": f"{self.forum_url}/{self.api_endpoint}/notifications",
            "groups": f"{self.forum_url}/{self.api_endpoint}/groups",

            "tags": f"{self.forum_url}/{self.api_endpoint}/tags",
        }


class FlarumUser(FlarumSession, dict):
    def __init__(self, extensions: Optional[List[ExtensionMixin]]=None, **kwargs):
        self.extensions = extensions

        if self.extensions:
            for extension in self.extensions:
                dependencies = extension.get_dependencies(extension) # type: dict

                hard = dependencies.get("hard", None)
                soft = dependencies.get("soft", None)

                if hard and len(hard) > 0:
                    for hard_dependency in hard:
                        if hard_dependency not in self.extensions:
                            raise MissingExtensionError(f'`{extension}` hardly depends on `{hard_dependency}`. Please, include that extension too in your extension list.')

                extension.mixin(extension)

            for extension in self.extensions:
                if soft and len(soft) > 0:
                    for soft_dependency in soft:
                        if soft_dependency not in self.extensions:
                            warnings.warn(f'`{extension}` softly depends on `{soft_dependency}`. Some features might be unavailable.', MissingExtensionWarning)


        super().__init__(**kwargs)
        __json = self._fetch_user_data()


        if __json:
            self.data = MyUser(user=self, _fetched_data=dict(data=__json))


    def _fetch_user_data(self):
        """
            Fetches your user's JSON data.
        """

        filter = Filter(query=self.username, limit=1)
        raw = self.session.get(f"{self.api_urls['users']}", params=filter.to_dict)
        json = parse_request(raw)

        for possible_user in json.get("data", {}):
            if possible_user.get("attributes", {}).get("username", None) == self.username:
                return possible_user

        return None


    def _update_user_data(self, new_data: dict):
        if new_data.get("data", {}).get("id", None) == self.data.id:
            self.data = MyUser(user=self, _fetched_data=new_data)

            return self

        else:
            return self


    def get_forum_data(self):
        raw = self.session.get(f"{self.api_urls['base']}")
        json = parse_request(raw)

        return Forum(user=self, _fetched_data=json)


    def get_discussion_by_id(self, id: int):
        raw = self.session.get(f"{self.api_urls['discussions']}/{id}")
        json = parse_request(raw)

        return Discussion(user=self, _fetched_data=json)


    def get_post_by_id(self, id: int):
        raw = self.session.get(f"{self.api_urls['posts']}/{id}")
        json = parse_request(raw)

        return Post(user=self, _fetched_data=json)


    def all_discussions(self, filter: Filter=None):
        """
            Obtains all discussions from specific page by using `filter`.
        """


        if filter:
            raw = self.session.get(f"{self.api_urls['discussions']}", params=filter.to_dict)

        else:
            raw = self.session.get(f"{self.api_urls['discussions']}")

        json = parse_request(raw)

        return Discussions(user=self, _fetched_data=json)


    def all_posts(self, filter: Filter=None):
        """
            Obtains all posts from specific page by using `filter`.
        """


        if filter:
            raw = self.session.get(f"{self.api_urls['posts']}", params=filter.to_dict)

        else:
            raw = self.session.get(f"{self.api_urls['posts']}")

        json = parse_request(raw)

        return Posts(user=self, _fetched_data=json)


    def all_users(self, filter: Filter=None):
        """
            Obtains all users from specific page by using `filter`.
        """


        if filter:
            raw = self.session.get(f"{self.api_urls['users']}", params=filter.to_dict)

        else:
            raw = self.session.get(f"{self.api_urls['users']}")

        json = parse_request(raw)

        return Users(user=self, _fetched_data=json)


    def get_notifications(self, filter: Optional[Filter]=None) -> Notifications:
        """
            Obtains all notifications of your user.
        """

        if filter:
            raw = self.session.get(f"{self.api_urls['notifications']}", params=filter.to_dict)
        
        else:
            raw = self.session.get(f"{self.api_urls['notifications']}")


        json = parse_request(raw)

        return Notifications(user=self, _fetched_data=json)


    def mark_all_discussions_as_read(self, at: datetime=datetime.now()):
        post_data = {
            "data": {
                "type": "users",
                "id": self.data.id,
                "attributes": {
                    "markedAllAsReadAt": datetime_to_flarum(at)
                }
            }
        }

        raw = self.session.patch(f"{self.api_urls['users']}/{self.data.id}", json=post_data)
        parse_request(raw)

        return True


    def mark_all_notifications_as_read(self):
        raw = self.session.post(f"{self.api_urls['notifications']}/read")
        parse_request(raw)

        return True


    def get_groups(self):
        raw = self.session.get(f"{self.api_urls['groups']}")
        json = parse_request(raw)

        return Groups(user=self, _fetched_data=json)


    def get_user_by_id(self, id: int):
        raw = self.session.get(f"{self.api_urls['users']}/{id}")
        json = parse_request(raw)

        return User(user=self, _fetched_data=json)


    def update_user_info(self, user: Optional[AnyUser]=None, new_username: Optional[str]=None, groups: Optional[list]=None):
        id = user.id if user else self.data.id

        patch_data = {
            "data": {
                "type": "users",
                "id": id,
                "attributes": {},
                "relationships": {}
            }
        } # type: Dict[str, Dict[str, Union[Any, dict]]]


        if new_username:
            patch_data['attributes']['username'] = new_username

        if groups:
            patch_data['data']['relationships'].update({"groups": {"data": groups}})


        raw = self.session.patch(f"{self.api_urls['users']}/{id}", json=patch_data)
        json = parse_request(raw)

        return self._update_user_data(new_data=dict(data=json))


    def send_password_reset_email(self):
        patch_data = { "email": self.data.email }

        raw = self.session.patch(f"{self.api_urls['users']}/{self.data.id}", json=patch_data)
        json = parse_request(raw)

        return self._update_user_data(new_data=dict(data=json))


    def update_preferences(self, notification: str, type: str=Literal['alert', 'email'], user: Optional[AnyUser]=None, state: bool=True):
        id = user.id if user else self.data.id

        patch_data = {
            "data": {
                "type": "users",
                "id": id,
                "attributes": {
                    "preferences": {
                        f"notify_{notification}_{type}": state
                    }
                }
            }
        }

        raw = self.session.patch(f"{self.api_urls['users']}/{id}", json=patch_data)
        json = parse_request(raw)

        return self._update_user_data(new_data=json)


    def change_email(self, new_email: str, password_confirmation: str, user: Optional[AnyUser]=None):
        id = user.id if user else self.data.id

        patch_data = {
            "data": {
                "type": "users",
                "id": id,
                "attributes": {
                    "email": new_email
                }
            },
            "meta": {
                "password": password_confirmation
            }
        }

        raw = self.session.patch(f"{self.api_urls['users']}/{id}", json=patch_data)
        json = parse_request(raw)

        return self._update_user_data(new_data=json)


    def upload_user_avatar(self, file: BinaryIO, user: Optional[AnyUser]=None, file_name: str="avatar", file_type: Literal['image/png', 'image/jpeg', 'image/gif']="image/png"):
        raw = self.session.post(url=f"{self.api_urls['users']}/{user.id if user else self.data.id}/avatar", files={ "avatar": (file_name, file, file_type) })

        json = parse_request(raw)

        return self._update_user_data(new_data=json)


    def remove_user_avatar(self, user: Optional[AnyUser]=None):
        raw = self.session.delete(url=f"{self.api_urls['users']}/{user.id if user else self.data.id}/avatar")
        json = parse_request(raw)

        return self._update_user_data(new_data=json)



from typing import List, Optional

from datetime import datetime

from .. import ExtensionMixin
from ...session import FlarumUser

from ...flarum.core.forum import Forum
from ...flarum.core.discussions import DiscussionFromBulk
from ...datetime_conversions import flarum_to_datetime
from ...error_handler import parse_request


AUTHOR = 'flarum'
NAME = 'tags'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class Tag(dict):
    """
        A Flarum tag.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user
        super().__init__(_fetched_data)


    @property
    def data(self) -> dict:
        return self.get("data", {})


    @property
    def type(self) -> Optional[str]:
        return self.data.get("type", None)


    @property
    def id(self) -> Optional[int]:
        return self.data.get("id", None)


    @property
    def attributes(self) -> dict:
        return self.data.get("attributes", {})


    @property
    def name(self) -> Optional[str]:
        return self.attributes.get("name", None)


    @property
    def description(self) -> Optional[str]:
        return self.attributes.get("description", None)


    @property
    def slug(self) -> Optional[str]:
        return self.attributes.get("slug", None)


    @property
    def color(self) -> str:
        return self.attributes.get("color", "")


    @property
    def backgroundUrl(self) -> Optional[str]:
        return self.attributes.get("backgroundUrl", None)


    @property
    def backgroundMode(self) -> Optional[str]:
        return self.attributes.get("backgroundMode", None)


    @property
    def icon(self) -> Optional[str]:
        return self.attributes.get("icon", None)


    @property
    def discussionCount(self) -> Optional[int]:
        return self.attributes.get("discussionCount", None)


    @property
    def discussionCount(self) -> Optional[int]:
        return self.attributes.get("discussionCount", None)


    @property
    def position(self) -> Optional[int]:
        return self.attributes.get("position", None)


    @property
    def defaultSort(self) -> Optional[str]:
        return self.attributes.get("defaultSort", None)


    @property
    def isChild(self) -> bool:
        return self.attributes.get("isChild", False)


    @property
    def isHidden(self) -> bool:
        return self.attributes.get("isHidden", False)


    @property
    def lastPostedAt(self) -> Optional[datetime]:
        raw = self.attributes.get("lastPostedAt", None)
        return flarum_to_datetime(raw)


    @property
    def canStartDiscussion(self) -> bool:
        return self.attributes.get("canStartDiscussion", False)


    @property
    def canAddToDiscussion(self) -> bool:
        return self.attributes.get("canAddToDiscussion", False)


    @property
    def subscription(self) -> Optional[str]:
        return self.attributes.get("subscription", None)


    @property
    def template(self) -> str:
        return self.attributes.get("template", "")


    @property
    def relationships(self) -> dict:
        return self.data.get("relationships", {})


    def get_parent_tag(self, force: bool=False):
        if not self.isChild and not force:
            raise TypeError(f'{self.name} is not a children, and therefore does not have a parent tag. Use `force = True` to bypass this error.')

        id = self.relationships.get("parent", {}).get("data", {}).get("id", None) # type: int
        raw = self.user.session.get(f"{self.user.api_urls['tags']}")
        json = parse_request(raw)

        for raw_tag in json.get("data", [{}]):
            if (raw_tag.get("id", None) == id) and (raw_tag.get("type", None) == 'tags'):
                return Tag(user=self.user, _fetched_data=dict(data=raw_tag))

        return None


    def __restrict_or_unrestrict_permissions(self, restrict: bool=True):
        post_data = {
            "data": {
                "type": "tags",
                "id": "1",
                "attributes": {
                    "isRestricted": restrict
                }
            }
        }

        raw = self.user.session.post(f"{self.user.api_urls['tags']}/{self.id}", json=post_data)
        json = parse_request(raw)

        return Tag(user=self.user, _fetched_data=json)


    def restrict_permissions(self):
        return self.__restrict_or_unrestrict_permissions(restrict=True)


    def unrestrict_permissions(self):
        return self.__restrict_or_unrestrict_permissions(restrict=False)



class TagsForumMixin(Forum):
    @property
    def canBypassTagCounts(self) -> bool:
        return self.attributes.get("canBypassTagCounts", False)


    @property
    def minPrimaryTags(self) -> Optional[int]:
        raw = self.attributes.get("minPrimaryTags", None)

        if raw:
            return int(raw)


    @property
    def maxPrimaryTags(self) -> Optional[int]:
        raw = self.attributes.get("maxPrimaryTags", None)

        if raw:
            return int(raw)


    @property
    def minSecondaryTags(self) -> Optional[int]:
        raw = self.attributes.get("minSecondaryTags", None)

        if raw:
            return int(raw)


    @property
    def maxSecondaryTags(self) -> Optional[int]:
        raw = self.attributes.get("maxSecondaryTags", None)

        if raw:
            return int(raw)



class TagsDiscussionMixin(DiscussionFromBulk):
    @property
    def canTag(self) -> bool:
        return self.attributes.get("canTag", False)


    def get_tags(self) -> List[Tag]:
        all_tags = [] # type: List[Tag]
        seen = set()
        tags = self.relationships.get("tags", {}).get("data", [{}]) # type: List[dict]

        for raw_tag in tags:
            id = raw_tag.get("id", None)

            for possible_tag in self._parent_included:
                if possible_tag and (possible_tag.get("type", None) == 'tags') and (possible_tag.get("id", None) == id) and (possible_tag.get("id", None) not in seen):
                    tag = Tag(user=self.user, _fetched_data=dict(data=possible_tag))
                    all_tags.append(tag)

                    seen.add(tag.id)
                    continue

        return all_tags



class TagsExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, Forum, TagsForumMixin)
        super().mixin(self, DiscussionFromBulk, TagsDiscussionMixin)



from .. import ExtensionMixin

from ...flarum.core.discussions import DiscussionFromBulk


AUTHOR = 'blomstra'
NAME = 'realtime'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []


class RealtimeDiscussionMixin(DiscussionFromBulk):
    @property
    def canViewWhoTypes(self) -> bool:
        return self.attributes.get("canViewWhoTypes", False)



class RealtimeExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, DiscussionFromBulk, RealtimeDiscussionMixin)



import sys
sys.path.append('.')

import pyflarum
import setuptools


with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()


setuptools.setup(
    name = 'pyFlarum',
    packages = setuptools.find_packages(),

    long_description=long_description,
    long_description_content_type='text/markdown',

    version = pyflarum.__version__,
    license=pyflarum.__license__,
    description = pyflarum.__description__,
    keywords = ['Flarum', 'forum software', 'api', 'Flarum api', 'bot', 'userbot', 'database', 'flarum database', 'server', 'client'],

    author = pyflarum.__author__,
    author_email = pyflarum.__email__,

    url = 'https://github.com/CWKevo/pyflarum',

    install_requires=[
        'requests'
    ],

    classifiers=[
        f'Development Status :: {pyflarum.__status__}',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



from typing import List

from . import ExtensionMixin
from ..session import FlarumUser

import re


AUTHOR = 'skevo'
NAME = 'commands'
ID = f"{AUTHOR}-{NAME}"


SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []


class CommandsFlarumUserMixin(FlarumUser):
    def is_mentioned_in(self, string: str):
        match = re.search(rf'@"({self.username})"#[a-z0-9]+', string)

        if match:
            return True

        else:
            return False


    def parse_as_command(self, string: str, is_mentioned: bool=True, split_at: str=' '):
        """
            Parses a command from a string (e. g.: post's content). The result is list of arguments.

            ### Example:
            ```python
                user.parse_command()
            ```
        """

        if self.is_mentioned_in(string) and is_mentioned:
            match = re.search(rf'@"{self.username}"#[a-z0-9]+\s(.+)', string)

            if match:
                group = match.group(1)

                if group:
                    parsed = group.split(sep=split_at) # type: List[str]

                    return parsed

        return string.split(sep=split_at)


class CommandsExtension(ExtensionMixin, CommandsFlarumUserMixin):
    def __init__(self):
        self.name = NAME
        self.author = AUTHOR
        self.id = ID

    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, FlarumUser, CommandsFlarumUserMixin)



from typing import Literal

from .. import ExtensionMixin

from ...flarum.core.discussions import Discussion, DiscussionFromNotification, DiscussionFromBulk

from ...error_handler import parse_request


AUTHOR = 'flarum'
NAME = 'subscriptions'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []


class SubscriptionsDiscussionFromNotificationMixin(DiscussionFromNotification):
    def __change_subscription_state(self, state: Literal['following', 'ignoring', None]):
        post_data = {
            "data": {
                "type": "discussions",
                "id": "15",
                "attributes": {
                    "subscription": state
                }
            }
        }

        raw = self.user.session.patch(f"{self.user.api_urls['discussions']}/{self.id}", json=post_data)
        json = parse_request(raw)

        return Discussion(user=self.user, _fetched_data=json)
    

    def follow(self):
        return self.__change_subscription_state(state='following')


    def unfollow(self):
        return self.__change_subscription_state(state=None)


    def ignore(self):
        return self.__change_subscription_state(state='ignore')


class SubscriptionsDiscussionFromBulkMixin(DiscussionFromBulk, SubscriptionsDiscussionFromNotificationMixin):
    @property
    def subscription(self) -> bool:
        return self.attributes.get("subscription", None)



class SubscriptionsExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, DiscussionFromNotification, SubscriptionsDiscussionFromNotificationMixin)
        super().mixin(self, DiscussionFromBulk, SubscriptionsDiscussionFromBulkMixin)



from .. import ExtensionMixin

from ...flarum.core.forum import Forum


AUTHOR = "flarum"
NAME = "markdown"
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class ForumMixin(Forum):
    @property
    def markdown_mdarea(self) -> bool:
        return self.attributes.get("flarum-markdown.mdarea", False)



class ExampleExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, Forum, ForumMixin)
